;;; DO NOT MODIFY THIS FILE
(if (featurep 'semantic-autoloads) (error "Already loaded"))

;;;### (autoloads nil "_pkg" "semantic/_pkg.el")

(package-provide 'semantic :version 1.06 :type 'regular)

;;;***

;;;### (autoloads (semantic-create-imenu-index) "semantic-imenu" "semantic/semantic-imenu.el")

(autoload 'semantic-create-imenu-index "semantic-imenu" "\
Create an imenu index for any buffer which supports Semantic.
Uses the output of the Semantic Bovinator to create the index.
Optional argument STREAM is an optional stream of tokens used to create menus." nil nil)

;;;***

;;;### (autoloads (semantic-minor-mode) "semantic-mode" "semantic/semantic-mode.el")

(autoload 'semantic-minor-mode "semantic-mode" "\
Toggle semantic minor mode, a mode for tracking buffer parsing.
With prefix ARG, turn semantic minor mode on iff ARG is positive.

When in Semantic minor mode, the buffer will be parsed, and reparsed
during editing in small chunks, during idle time.

Semantic minor mode also allows modification of text properties on
tokens, highlighting them, or making functions read only.

\\<semantic-mode-map>
\\{semantic-mode-map}" t nil)

;;;***

;;;### (autoloads (semantic-bovinate-toplevel) "semantic" "semantic/semantic.el")

(autoload 'semantic-bovinate-toplevel "semantic" "\
Bovinate the entire current buffer.
If the optional argument CHECKCACHE is non-nil, then flush the cache iff
there has been a size change." nil nil)

;;;***

;;;### (autoloads (senator-word-search-backward senator-re-search-backward senator-search-backward senator-word-search-forward senator-re-search-forward senator-search-forward senator-jump senator-previous-token senator-next-token) "senator" "semantic/senator.el")

(autoload 'senator-next-token "senator" "\
Navigate to the next semantic token.
Return the semantic token or nil if at end of buffer." t nil)

(autoload 'senator-previous-token "senator" "\
Navigate to the previous semantic token.
Return the semantic token or nil if at beginning of buffer." t nil)

(autoload 'senator-jump "senator" "\
Jump to the semantic symbol SYM.
If called interactively and a prefix argument is supplied jump in the
local type's context (see function `senator-current-type-context')." t nil)

(autoload 'senator-search-forward "senator" "\
Search semantic tokens forward from point for string WHAT.
Set point to the end of the occurrence found, and return point.  See
`search-forward' for details and the meaning of BOUND NOERROR and
COUNT.  BOUND and COUNT are just ignored in the current
implementation." t nil)

(autoload 'senator-re-search-forward "senator" "\
Search semantic tokens forward from point for regexp WHAT.
Set point to the end of the occurrence found, and return point.  See
`re-search-forward' for details and the meaning of BOUND NOERROR and
COUNT.  BOUND and COUNT are just ignored in the current
implementation." t nil)

(autoload 'senator-word-search-forward "senator" "\
Search semantic tokens forward from point for word WHAT.
Set point to the end of the occurrence found, and return point.  See
`word-search-forward' for details and the meaning of BOUND NOERROR and
COUNT.  BOUND and COUNT are just ignored in the current
implementation." t nil)

(autoload 'senator-search-backward "senator" "\
Search semantic tokens backward from point for string WHAT.
Set point to the beginning of the occurrence found, and return point.
See `search-backward' for details and the meaning of BOUND NOERROR and
COUNT.  BOUND and COUNT are just ignored in the current
implementation." t nil)

(autoload 'senator-re-search-backward "senator" "\
Search semantic tokens backward from point for regexp WHAT.
Set point to the beginning of the occurrence found, and return point.
See `re-search-backward' for details and the meaning of BOUND NOERROR
and COUNT.  BOUND and COUNT are just ignored in the current
implementation." t nil)

(autoload 'senator-word-search-backward "senator" "\
Search semantic tokens backward from point for word WHAT.
Set point to the beginning of the occurrence found, and return point.
See `word-search-backward' for details and the meaning of BOUND
NOERROR and COUNT.  BOUND and COUNT are just ignored in the current
implementation." t nil)

;;;***

(provide 'semantic-autoloads)
